/*  "Expressive" compiler
 *
 *  Code emitting abstractions
 *  Nikolay Mihaylov
 *
 *  $Id: Code.java,v 1.5 2003/06/04 13:34:26 mihaylov Exp $
 */

package expressive;

import expressive.Tree.*;

import java.util.Vector;
import java.util.HashMap;

/**
 * This class implements code emitting abstractions for the EVM
 */
public class Code implements expressive.evm.Instructions {

    //########################################################################
    // Nested classes

    /** List of locations that point to a label */
    protected static final class RefList {
	public final int ref;
	public final RefList next;
	public RefList(int ref, RefList next) {
	    this.ref = ref;
	    this.next = next;
	}
    }

    /**
     * Represents labels to which branch instructions can point
     */
    public final class Label {
	private static final int UNDEFINED = -1;
	private int pos = UNDEFINED;;
	private RefList refs = null;

	/** Has the label already been defined? */
	public boolean isDefined() {
	    return pos != UNDEFINED;
	}

	/** Add a reference to the label. */
	public void addRef(int to) {
	    if (pos != UNDEFINED)
		code.set(to, new Integer(pos));
	    else
		refs = new RefList(to, refs);
	}

	/** Set the label at the current position in the code. */
	public void set() {
	    int to = code.size();
	    assert pos == UNDEFINED :
		"This label has already been set to " + pos;
	    for (RefList rl = refs; rl != null; rl = rl.next)
		code.set(rl.ref, new Integer(to));
	    pos = to;
	    refs = null;
	}
    }

    //########################################################################
    // Instance variables

    private final Vector/*<int>*/ code = new Vector();

    private final Vector/*<String>*/ strings = new Vector();

    // Map strings to constant pool indices
    private final HashMap/*<String, int>*/ stringIndices = new HashMap();

    private int memSize = 0;


    //########################################################################
    // Instrumental methods

    /** Return the value of the Program Counter. */
    public int pc() { return code.size(); }

    /** Return the size of the data memory area. */
    public int getMemSize() { return memSize; }

    /** Set the size of the data memory area. */
    public void setMemSize(int ms) { memSize = ms; }

    /** Skip one position in the code sequence.
     *  Useful where the operand of an instruction
     *  will be determined later - goto, jz, jnz
     */
    private int skip() {
	int pc = pc();
	code.add(null);
	return pc;
    }

    //########################################################################
    // Constructors

    public Code() {}

    //########################################################################
    // Methods

    /** Emit instruction without operands. */
    public Code emit(int instr) {
	code.add(new Integer(instr));
	return this;
    }

    /** Emit instruction with integer operand. */
    public Code emit(int instr, int arg) {
	emit(instr);
	emit(arg);
	return this;
    }

    /** Emit instruction with string operand (PRINTS). */
    public Code emit(int instr, String str) {
	Integer index = (Integer) stringIndices.get(str);
	if (index == null) {
	    index = new Integer(strings.size());
	    strings.add(str);
	    stringIndices.put(str, index);
	}
	emit(instr);
	emit(index.intValue());
	return this;
    }

    /** Emit instruction with label operand (GOTO, JZ, JNZ). */
    public Code emit(int instr, Label label) {
	emit(instr);
	label.addRef(skip());
	return this;
    }

    /** Emit memory acces instruction (STORE). */
    public Code emit(int instr, Symbol sym) {
	return emit(instr, declareVar(sym));
    }

    /** Emit code sequence to negate the value at the top of the stack. */
    public Code emitNegate() {
	return emit(PUSH, 0).emit(CMPE);
    }

    /** Generate fresh label. */
    public Label getLabel() {
	return new Label();
    }

    /** Mark the label to point to the current location in the code. */
    public Label markLabel(Label label) {
	label.set();
	return label;
    }

    /** If necessary, declare fresh variable corresponiding to the symbol.
	Return the memory location of the symbol.
    */
    public int declareVar(Symbol sym) {
	if (sym.getOffset() < 0)
	    sym.setOffset(memSize++);
	return sym.getOffset();
    }

    /** Create new virtual machine for the generated code. */
    public expressive.evm.EVM getVM() {
	int[] instrs = new int[code.size()];
	for (int i = 0; i < instrs.length; i++) {
	    instrs[i] = ((Integer)code.get(i)).intValue();
	}
	String[] strs = (String[])strings.toArray(new String[strings.size()]);
	return new expressive.evm.EVM(instrs, strs, memSize);
    }

    //########################################################################

}
