/*  "Expressive" compiler
 *
 *  Pretty printer class
 *  Nikolay Mihaylov
 *
 *  $Id: Printer.java,v 1.7 2003/06/02 16:50:14 mihaylov Exp $
 */

package expressive;

import java.io.PrintWriter;

import expressive.Tree.*;

public class Printer implements OpCodes, Visitor {
    //########################################################################
    // Instance variable

    /** The output stream */
    private final PrintWriter out;

    /** The indentation step */
    private String step;

    /** The indentation level */
    private int level;

    //########################################################################
    // Constructors

    public Printer(PrintWriter out) {
        this.out = out;
        this.step = "  ";
        this.level = 0;
    }

    //########################################################################
    // Printing methods

    public Printer print(Tree tree) {
	if (tree == null) print("<EMPTY>");
        else tree.apply(this);
        return this;
    }

    public Printer print(Tree[] trees) {
	if (trees.length == 0) {
	    print("{}");
	} else {
	    print("{");
	    indent();
	    println();
	    for(int i = 0; i < trees.length; i++) {
		print(trees[i]).print(".");
		if (i == (trees.length - 1) ) {
		    undent();
		}
		println();
	    }
	    print("}");
	}
	return this;
    }

    public Printer print(int value) {
        return print(String.valueOf(value));
    }

    public Printer print(String value) {
        out.print(value);
        return this;
    }

    public Printer println() {
        out.println();
        for (int i = 0; i < level; i++) out.print(step);
        return this;
    }

    public Printer indent() {
        level++;
        return this;
    }

    public Printer undent() {
        level--;
        return this;
    }

    //########################################################################
    // case methods


    public void caseProg(Prog tree) {
	for (int i = 0; i < tree.body.length; i++)
	    print(tree.body[i]).print(".").println();
    }

    public void caseIO(IO tree) {
	switch (tree.op) {
	case PRINTNL: print("printnl"); break;
	case PRINT: print("print ").print(tree.expr); break;
	case READINT: print("readint ").print(tree.expr); break;
	case READBOOL: print("readbool ").print(tree.expr); break;
	}
    }

    public void caseAssign(Assign tree) {
	print(tree.name).print(" := ").print(tree.rhs);
    }

    public void caseLiteral(Literal tree) {
	if (tree.str == null) {
	    print(tree.value);
	} else {
	    print("\"").print(tree.str).print("\"");
	}
    }

    public void caseIdent(Ident tree) {
	print(tree.name);
    }

    public void caseOp(Op tree) {
	switch (tree.op) {
	case TRUE:  print("true"); break;
	case FALSE: print("false"); break;
	case NOT:
	    print("!");
	    if (tree.left != null) print(tree.left);
	    else if (tree.right != null) print(tree.right);
	    else print("<NO OPERAND>");
	    break;
	default:
	    print("(").print(tree.left);
	    switch (tree.op) {
	    case EQ: print(" = "); break;
	    case NE: print(" != "); break;
	    case LT: print(" < "); break;
	    case LE: print(" <= "); break;
	    case GT: print(" > "); break;
	    case GE: print(" >= "); break;
	    case OR: print(" | "); break;
	    case AND: print(" & "); break;
	    case PLUS: print(" + "); break;
	    case MINUS: print(" - "); break;
	    case TIMES: print(" * "); break;
	    case DIV: print(" / "); break;
	    case REM: print(" % "); break;
	    case POW: print(" ^ "); break;
	    default:
		print(" <UNKNOWN> ");
	    }
	    print(tree.right).print(")");
	}
    }

    public void caseIf(If tree) {
	print("if ").print(tree.cond).print(" ").print(tree.thenp);
	if (tree.elsep != null) {
	    println().print("else ").print(tree.elsep);
	}
    }

    public void caseWhile(While tree) {
	print("while ").print(tree.cond).print(" ").print(tree.body);
    }
}
