/*  "Expressive" compiler
 *
 *  Class that comprises the abstract syntax tree
 *  Nikolay Mihaylov
 *
 *  $Id: Tree.java,v 1.4 2003/05/20 09:57:11 mihaylov Exp $
 */

package expressive;

import java.util.List;

/**
 * Classes comprising the Abstract Syntax Tree
 */
public abstract class Tree {

    //########################################################################
    // Static members

    /** Empty array of statements. */
    static final public Stat[] NO_STATS = new Stat[0];

    /** Converts a list into an array of statements. */
    public static Stat[] toArray(List list) {
        return (Stat[])list.toArray(new Stat[list.size()]);
    }

    //########################################################################
    // Instance variables

    /** The type of this tree node (assigned during type analysis). */
    public Type type;

    /** Position in the sourcefile associated with this tree node. */
    public final int pos;

    //########################################################################
    // Constructors

    public Tree(int pos) {
	this.pos = pos;
    }

    //########################################################################

    /** Apllies a visitor to the tree node. */
    abstract public void apply(Visitor v);

    //########################################################################
    // Abstract superclasses for statements and expressions

    /** Abstract superclass for statements. */
    public static abstract class Stat extends Tree {
	public Stat(int pos) { super(pos); }
    }

    /** Abstract superclass for expressions. */
    public static abstract class Expr extends Tree {
	public Expr(int pos) { super(pos); }
    }

    //########################################################################
    // Tree node classes

    /** Tree node representing a program. */
    public static class Prog extends Tree {
	/** Statements comprising the body of the program. */
	public final Stat[] body;
	public Prog(Stat[] body) {
	    super(Position.UNDEFINED);
	    this.body = body;
	}
	public void apply(Visitor v) {
	    v.caseProg(this);
	}
    }

    /**
     * Tree node representing I/O statement:
     * print, printnl, readint, readbool.
     */
    public static class IO extends Stat {
	/** Opcode of the I/O operation. */
	public final int op;
	/** Argument of the operation (optional).*/
	public final Expr expr;
	public IO(int pos, int op, Expr expr) {
	    super(pos);
	    this.op = op;
	    this.expr = expr;
	}
	public void apply(Visitor v) {
	    v.caseIO(this);
	}
    }

    /** Tree node representing assignment statement. */
    public static class Assign extends Stat {
	/** Name of the identifier on the left-hand side. */
	public final String name;
	/** Right-hand side of the assignment. */
	public final Expr rhs;
	/** The symbol of the identifier. */
	public Symbol symbol;
	public Assign(int pos, String name, Expr rhs) {
	    super(pos);
	    this.name = name;
	    this.rhs = rhs;
	}
	public void apply(Visitor v) {
	    v.caseAssign(this);
	}
    }

    /** Tree node representing if E then Ss Else Ss. */
    public static class If extends Stat {
	/** Condition. */
	public final Expr cond;
	/** Then part of the statement. */
	public final Stat[] thenp;
	/** Else part of the statement. */
	public final Stat[] elsep;
	public If(int pos, Expr cond, Stat[] thenp, Stat[] elsep) {
	    super(pos);
	    this.cond = cond;
	    this.thenp = thenp;
	    this.elsep = elsep;
	}
	public void apply(Visitor v) {
	    v.caseIf(this);
	}
    }

    /** Tree node representing while E do Ss. */
    public static class While extends Stat {
	/** Condition of the loop. */
	public final Expr cond;
	/** Body of the loop. */
	public final Stat[] body;
	public While(int pos, Expr cond, Stat[] body) {
	    super(pos);
	    this.cond = cond;
	    this.body = body;
	}
	public void apply(Visitor v) {
	    v.caseWhile(this);
	}
    }

    /** Tree node representing operations. */
    public static class Op extends Expr {
	/** Opcode of the operation. */
	public final int op;
	/** Left operand (optional). */
	public final Expr left;
	/** Left operand (optional). */
	public final Expr right;
	public Op(int pos, Expr left, int op, Expr right) {
	    super(pos);
	    this.left = left;
	    this.right = right;
	    this.op = op;
	}
	public void apply(Visitor v) {
	    v.caseOp(this);
	}
    }

    /** Tree node representing identifiers. */
    public static class Ident extends Expr {
	/** Name of the identifier. */
	public final String name;
	/** Symbol representing the identifier (assigned during name analysis)*/
	public Symbol symbol;
	public Ident(int pos, String name) {
	    super(pos);
	    this.name = name;
	}
	public void apply(Visitor v) {
	    v.caseIdent(this);
	}
    }

    /** Tree node representing literals - ints and strings. */
    public static class Literal extends Expr {
	/** Value of the literal (if it's integer). */
	public final int value;
	/** Value of a string literal, null otherwise */
	public final String str;
	public Literal(int pos, int value) {
	    super(pos);
	    this.value = value;
	    this.str  = null;
	}
	public Literal(int pos, String str) {
	    super(pos);
	    this.value = 0;
	    this.str = str;
	}
	public void apply(Visitor v) {
	    v.caseLiteral(this);
	}
    }

    //########################################################################

}
