/* Evm.java
 * (c) 2003 Erik Stenman. 
 *
 * Beware - this is my first Java program ever,
 *
 * This class implements a simple virtual machine designed for the
 * language Expressive used in the course Compilation A
 * http://lampwww.epfl.ch/teaching/CompilationA/2003/html/ (Since
 * Expressive has no concept of functions, the machine also lacks
 * instructions for function calls and there is no return
 * instruction.)
 *
 * EVM is a stack machine. For description of the instruction set
 * look at (the documentation for) evm/Instructions.java
 *
 * $Id: EVM.java,v 1.4 2003/06/05 07:40:59 mihaylov Exp $
 *
 */

package expressive.evm;

import java.io.*;

public final class EVM implements Instructions {

    //########################################################################
    // Instance variables

    /** The data stack - grows as needed. */
    protected int[] stack = new int[256];

    /** Program counter - points to next instruction in code. */
    protected int PC = 0;

    /** Stack pointer points to the element at the top of the stack. */
    protected int SP = -1;

    /** User input. */
    protected BufferedReader in =
	new BufferedReader(new InputStreamReader(System.in));

    /** Code to be executed by the VM. */
    protected final int[] code;

    /** Data memory area of the VM. */
    protected final int[] data;

    /** Pool of strings. */
    protected final String[] strings;

    //########################################################################
    // Constructors.

    /** 
     * EVM constructor
     * @param code Code to be executed by the VM
     * @param strings Pool of strings used by the program
     * @param memSize Size of the data memory area
     */
    public EVM(int[] code, String[] strings, int memSize) {
	this.code = code;
	this.strings = strings;
	this.data = new int[memSize];
    }

    //########################################################################
    // Methods

    /** Push an integer value on the stack. */
    protected void push(int x) {
	if (++SP >= stack.length) {
 	    int[] oldStack = stack;
 	    stack = new int[stack.length * 2];
 	    System.arraycopy(oldStack, 0, stack, 0, oldStack.length);
	}
	stack[SP] = x;
    }

    /** Get the value at the top of the stack*/
    protected int pop() {
	if (SP < 0) {
	    throw new RuntimeException
		("Stack underflow, @:" + PC + ": " +  instr(PC));
	}
	return stack[SP--];
    }

    /** Return the value of the instruction argument. */
    protected int arg() {
	if (PC >= code.length) {
	    throw new RuntimeException
		("No argument to push, @:" + PC + ": " +  instr(PC));
	}
	return code[PC++];
    }

    /** Return a string from the string pool. */
    protected String string(int at) {
	if (at < 0 || at >= strings.length) {
	    throw new RuntimeException
		("No string Nr: " + at + ", @:" + PC + ": " +  instr(PC));
	}
	return strings[at];
    }

    /** Return the value of a memory cell. */
    protected int data(int at) {
	if (at < 0 || at >= data.length) {
	    throw new RuntimeException
		("No memory cell: " + at + ", @:" + PC + ": " +  instr(PC));
	}
	return data[at];
    }	

    /** Compute the x ^ n. */
    protected static int pow(int x, int n) {
	if (x == 1) return 1;
	if (n < 0) return 0;
	if (n == 0) return 1;
	if (n % 2 == 1) return x * pow(x, n - 1);
	return pow(x * x, n / 2);
    }

    /** This is the main function of the machine, it executes the program
     *  by looping over the code, executing one instruction at a time.
     */
    public void run(PrintWriter out, boolean trace) {
	// Temps used for arithmetic.
	int x, y, r;
	//  The current instruction.
	int instr; 

	try {
	    while (true) {
		// Check that PC is valid.
		if (PC < 0 || PC >= code.length) {
		    throw new RuntimeException
			("Execution exited the code area, PC=" + PC);
		}
		// Show which instruction to execute if in debug mode.
		if (trace) printTraceInfo();

		// Execute the instruction pointed to by PC.
		instr = code[PC++];

		switch (instr) {

		case PUSH: // x - Push the argument. [b] -> [b,x]
		    push(arg());
		    break;

		case DUP: //  - Duplicate top of stack. [b,x] -> [b,x,x]
		    x = pop();
		    push(x);
		    push(x);
		    break;

		    // - Remove top of stack [b,x] -> [b]
		case POP: pop(); break;

		    // - Arithmetic and comparative operators. [b, x, y] -> [b, x OP y]
		case ADD:
		case SUB:
		case MUL:
		case DIV:
		case REM:
		case POW:
		case CMPL:
		case CMPE:
		case CMPLE:
		    y = pop();
		    x = pop();
		    switch (instr) {
		    case ADD: push(x + y); break;
		    case SUB: push(x - y); break;
		    case MUL: push(x * y); break;
		    case DIV: push(x / y); break;
		    case REM: push(x % y); break;
		    case POW: push(pow(x, y)); break;
		    case CMPL:  push((x <  y) ? 1 : 0); break;
		    case CMPLE: push((x <= y) ? 1 : 0); break;
		    case CMPE:  push((x == y) ? 1 : 0); break;
		    }
		    break;

		case STORE: // y - Store top of stack in mem: [b,x] -> [b], mem(y) = x
		    x = pop();
		    y = arg();
		    data(y);
		    data[y] = x;
		    break;

		case LOAD: // y -  Load top of stack from mem: [b] -> [b, mem(y)]
		    y = arg();
		    push(data(y));
		    break;

		case GOTO: // L - Jump to L: PC = L
		    PC = arg();
		    break;

		case JZ: // L - Jump on zero: [b,0] -> [b], PC = L | [b,x] -> [b]
		    x = pop();
		    y = arg();
		    if (x == 0) PC = y;
		    break;

		case JNZ: // L - Jump on not zero: [b,0] -> [b] | [b,x] -> [b], PC = L
		    x = pop();
		    y = arg();
		    if (x != 0) PC = y;
		    break;

		case PRINTI: //  - Print integer: [b,x] -> [b], printi x
		    x = pop();
		    out.print(x); out.flush();
		    break;

		case PRINTS: // s - Print string [b] -> [b], prints strings(s)
		    x = arg();
		    out.print(string(x)); out.flush();
		    break;

		case PRINTB: //  - Print bool: [b,x] -> [b], printb x
		    x = pop();
		    out.print(x != 0); out.flush();
		    break;

		case READI: // - Reads an int from the user: [b] -> [b,x]
		    while (true) {
			System.out.print("\n> ");		    
			try {
			    String input = in.readLine(); 
			    x = Integer.parseInt(input);
			    break;
			} catch (Exception e) {
			    System.out.print("(not an int)");
			}
		    }
		    push(x);
		    break;

		case READB: // - Reads a bool from the user: [b] -> [b,x]
		    while (true) {
			System.out.print("\nb> ");
			try {
			    String input = in.readLine(); 
			    if (input.equalsIgnoreCase("true")) x = 1;
			    else if (input.equalsIgnoreCase("false")) x = 0;
			    else {
				System.out.print("Enter 'true' or 'false'!");
				continue;
			    }
			    break;
			} catch (Exception e) {
			    System.out.print("Enter 'true' or 'false'!");
			}
		    }
		    push(x);
		    break;
		    
		case BREAK: // - Halts execution.
		    throw new RuntimeException();

		default:
		    System.err.println("Illeagal instruction " +
				       instr + " @ " +PC);
		    throw new RuntimeException();
		} // end switch
	    }// end while
	} catch (RuntimeException e) {
	    if (e.getMessage() != null)
	    System.err.println(e.getMessage());
	}
    }// end run


    /***/
    public String result() {
	if (SP < 0) return ""; else return "" + stack[SP];
    }

    /** Returns symbolic representation of the instruction at this address. */
    protected String instr(int at) {
	int arg = 0;;
	String arg_string;
	if (at + 1 < code.length) {
	    arg = code[at+1];
	    arg_string = /*pad(arg) */ "" +  arg;
	} else arg_string = "####";
	
	switch (code[at]) {
	case DUP:   return "dup"; 
	case POP:   return "pop"; 
	case ADD:   return "add"; 
	case SUB:   return "sub"; 
	case MUL:   return "mul"; 
	case DIV:   return "div"; 
	case REM:   return "rem"; 
	case POW:   return "pow"; 
	case CMPL:  return "cmpl"; 
	case CMPLE: return "cmple";  
	case CMPE:  return "cmpe";  
	case PRINTI:return "printi";  
	case PRINTB:return "printb";  
	case READI: return "readi";  
	case READB: return "readb";  
	case BREAK: return "break ";  
	case PUSH:  return "push     " + arg_string;
	case GOTO:  return "goto     " + arg_string;
	case STORE: return "store    " + arg_string;
	case LOAD:  return "load     " + arg_string;
	case JZ:    return "jz       " + arg_string;
	case JNZ:   return "jnz      " + arg_string;  
	case PRINTS:return "prints   \""   + getPrintable(strings[arg]) + "\"";
	default: return "UNKNOWN INSTRUCTION";
	}
    }

    /** Substites control characters with their escape sequence. */
    protected static String getPrintable(CharSequence chars) {
	StringBuffer str = new StringBuffer();
	for (int i = 0; i < chars.length(); i++) {
	    char ch = chars.charAt(i);
	    switch (ch) {
	    case '\n': str.append("\\n"); break;
	    case '\t': str.append("\\t"); break;
	    default: str.append(ch);
	    }
	}
	return str.toString();
    }

    /** Prints the code to the specified output. */
    public void printCode(PrintWriter out) {
	out.println("/* PC */     memsize  " + data.length);
	out.println();
	int s = code.length;
	int instr;
	for(int i =0; i<s; i++) {
	    instr = code[i];
	    out.println("/*" + pad(i) + i + "*/     " + instr(i));
	    if (instr >= PUSH) i++;
	}
    }

    /** Pads an integer with spaces on the left. */
    protected String pad(int i) {
	return (i<10)? "   " : ((i<100) ? "  " : ((i<1000) ? " " : ""));
    }

    protected  void printTraceInfo() {
	System.out.print(pad(PC) + PC + ": " + instr(PC) + " ");
	System.out.print("Top of stack: " + (SP < 0 ? "" : "" + stack[SP]) + " ");
	//System.out.print("Memory: " + data.toString());
	System.out.println();
    }

    //########################################################################

}
