/*  EVM Assembler
 *
 *  Program that reads assembler files and prints or executes them
 *
 *  Erik Stenman, Nikolay Mihaylov
 *
 *  $Id: Assembler.java,v 1.3 2003/06/04 13:34:06 mihaylov Exp $
 */

package expressive.evm.asm;

import java.io.InputStream;
import java.io.PrintWriter;
import java.util.HashMap;

import expressive.AbstractMain;
import expressive.Global;
import expressive.Code;

/**
 * Assembler for the EVM assembly language
 */
public class Assembler extends AbstractMain implements AsmInstructions {

    //########################################################################
    // Instance variables

    public boolean run = false;

    public boolean trace = false;

    protected final Code code = new Code();

    protected final HashMap/*<String,Label>*/ labels = new HashMap();

    //########################################################################
    // Instance methods

    /** Handle command line options and return
     *  the number of the consumed arguments. */
    public int handleOption(String[] args, int i) {
	if (args[i].equals("-run")) {
	    run = true;
	    return 1;
	} else if (args[i].equals("-trace")) {
	    trace = true;
	    return 1;
	} else return super.handleOption(args, i);
    }

    /** Print help info. */
    public void printHelp(java.io.PrintStream out) {
        printUsage(out);
        out.println("with the following options :");
        out.println("  -debug    Print the debug messages");
        out.println("  -run      Run the generated program if there were no errors");
        out.println("  -trace    Trace the execution of the program");
        out.println("  -? -help  Show this help");
    }

    /** Main program. */
    public void run(Global global, InputStream in, PrintWriter out) {
	final AsmScanner scanner = new AsmScanner(in);
	int memSize = 1024; // the default data memory size
	try {
	    Token token = scanner.next_token();
	    while (token.sym != EOF) {
		if (token.sym < have_arg) {
		    code.emit(token.sym);
		} else if (token.sym < have_string_arg) {
		    Token arg = scanner.next_token();
		    if (arg.sym == INT) {
			code.emit(token.sym, arg.ivalue);
		    } else if (arg.sym == IDENT) {
			code.emit(token.sym, getLabel(arg.svalue));
		    } else {
			global.error(arg.pos, "Integer or label argument expected!");
		    }
		} else if (token.sym == PRINTS) {
		    Token arg = scanner.next_token();
		    if (arg.sym == STRING) {
			assert arg.svalue != null : "String tokens need an svalue";
			code.emit(token.sym, arg.svalue);
		    } else {
			global.error(arg.pos, "String argument expected!");
		    }
		} else if (token.sym == IDENT) {
		    Token col = scanner.next_token();
		    if (col.sym != COLON)
			global.error(col.sym, "':' expected!");
		    assert token.svalue != null;
		    Code.Label l = getLabel(token.svalue);
		    if (!l.isDefined()) {
			code.markLabel(l);
		    } else {
			global.error(token.pos, "Label '" + token.svalue +
				     "' already defined!");
		    }
		} else if (token.sym == MEM_SIZE) {
		    Token arg = scanner.next_token();
		    if (arg.sym == INT) {
			memSize = arg.ivalue;
		    } else {
			global.error(arg.pos, "'memsize' directive expects " +
				     "integer argument!");
		    }
		} else {
		    global.error(token.pos, "Unknown token: " + token.sym);
		}
		token = scanner.next_token();
	    }
	} catch (java.io.IOException e) { e.printStackTrace(); }

	code.setMemSize(memSize);
	if (run && (global.errors() == 0)) {
	    code.getVM().run(out, trace);
	} else {
	    code.getVM().printCode(out);
	}
	out.println();
    }

    /** Returns a new label if no such exists. */
    protected Code.Label getLabel(String name) {
	Code.Label l = (Code.Label) labels.get(name);
	if (l == null) {
	    l = code.getLabel();
	    labels.put(name, l);
	}
	return l;
    }

    //########################################################################
    // Main method

    public static void main(String[] args) {
	new Assembler().run(args);
    }

    //########################################################################
}
